// Doo-Sabin Subdivision

// This program performs a Doo-Sabin subdivision of the mesh
// that is hard-coded into it. The object can be rotated with a trackball.
// Each time the "+" key is pressed, the subdivision level is increased by 1.

// It probably should be split into several files, but I find the process
// of defining the .h files distinctly tedious. Also, splitting a program into
// multiple files doesn't work anywhere near as well in C as in Java.
// But if you don't like one long file, feel free to split it yourself :)

// The implementation was intended to be as simple as possible but still
// grew to be a rather complex beast. I have a feeling that the problem
// shouldn't be this hard, but other people's solutions don't look any better
// than this one!

// Written for COMPSCI 715, Richard Lobb, July 2003.

#include <windows.h>
#include <gl/gl.h>
#include <gl/glu.h>
#include <gl/glut.h>
#include <assert.h>
#include "geometry.h"
#include "Trackball.h"

// C++ STL Headers

#include <algorithm>
#include <iostream>
#include <vector>

using namespace std;

typedef vector<int> intList;
typedef vector<CVec3df> PointList;
const float PI = 3.1415926;  // WHY DOESN'T MICROSOFT PROVIDE M_PI?!!


/*******************************************************
 * 
 * The Face class
 * ==============
 *
 * A face is just a list of vertex IDs, i.e. of integer indices
 * into a vertex table. 
 *
 *******************************************************/
class Face {
  public:
    intList vertices;      // Indices into vertex list
    Face() { vertices.clear(); }
    Face(int n, int indices[]) {
       vertices.clear();
       for (int i = 0; i < n; i++)
           addVertex(indices[i]);
    }
    void addVertex(int v) { vertices.push_back(v); }
    int size() const { return (int) vertices.size(); }
    int operator[](int index) const { return vertices[index]; }
    friend ostream& operator<<(ostream& s, const Face& vf);
    int indexToVertex(int vReqd) {
    // Searches for the required vertex (which is an integer vertex table index)
        for (int i = 0; i < (int) vertices.size(); i++)
           if (vertices[i] == vReqd) return i;
        return -1;  // Not found
    }
};

ostream& operator<<(ostream& s, const Face& f) {
   s << "(";
   for (int i = 0; i < f.size(); i++) {
      s << f[i];
      if (i != f.size() - 1) s << ",";
   }
   s << ")";
   return s;
}

typedef vector<Face> FaceList;

/*******************************************************
 * 
 * The Edge class
 * ==============
 *
 * A Edge is a pair of vertex IDs, i.e. of integer indices
 * into a vertex table, together with the ID (i.e. index into
 * a Face table) of the Face to the left of the edge when
 * going from the first vertex to the second.
 *
 * Note that each edge in the mesh has two Edge records,
 * one for each "direction of travel". Hence the two faces
 * sharing an edge are found by looking up the two Edge
 * records in the Edge table (which is an associative store,
 * implemented as an STL hash_set).
 *
 *******************************************************/
 
class Edge {
  public:
    // An edge is a pair of vertices (indices into vertex table).
    int v1, v2;
    int leftFace;  // The ID (index in face table) of the left face.
	Edge* next;    // Edges are linked as well as being in a table.
    Edge(int a, int b, int left) {
       v1 = a;
       v2 = b;
       leftFace = left;
    }
	Edge() { v1 = -1; } 	// An empty edge
	bool isVoid() {return v1 == -1;}
	int hash() { return v1 * 123457 ^ 3791 * v2; }  // A quick and dirty hash code
	friend bool operator==(const Edge e1, const Edge e2);
    friend ostream& operator<<(ostream& s, const Edge& e);
};

// Supporting routines for the Edge class
// --------------------------------------

ostream& operator<<(ostream& s, const Edge& e) {
   s << "(" << e.v1 << "," << e.v2 << ")";
   return s;
}

bool operator==(const Edge e1, const Edge e2) {
	return e1.v1 == e2.v1 && e1.v2 == e2.v2;
}


/*******************************************************
 * 
 * The EdgeTable class
 * ==================
 *
 * A EdgeTable is a hashtable of the edges in a mesh.
 * I originally implemented this as an STL hash_set, but the mingw
 * and Microsoft implementations of this are incompatible. GRRRRRR!
 *
 * As edges are inserted into the table they are chained for sequential access.
 *
 *******************************************************/

class EdgeTable {
  public:
	Edge* data;
	Edge* firstEdge;
	int tableSize;
	int numItems;
	EdgeTable(int maxNumEdges) {
		tableSize = 3 * maxNumEdges;
		data = new Edge[tableSize];
		numItems = 0;
		firstEdge = 0;
	}
	EdgeTable() { tableSize = 0; }   // Compiler wants this
	void insert(Edge e) {
		assert(numItems++ < 2 * tableSize / 3);
		int index = e.hash() % tableSize;
		if (index < 0) index += tableSize;
		assert (index >= 0);
		while (! data[index].isVoid()) {
			index = (++index) % tableSize;
		}
		data[index] = e;
		data[index].next = firstEdge;
		firstEdge = &data[index];
	}

	Edge* find(Edge e) {
		// Looks up e in the EdgeTable. This has two purposes.
		// (1) To determine if it has already been stored.
		// (2) To retrieve the leftFace field (which won't be
		//	   defined in e but is in the stored entry.)
		// I probably should use a hash_map equivalent for this
		// but I didn't. So there.
		int index = e.hash() % tableSize;
		while (! data[index].isVoid()) {
			if (data[index] == e) return &data[index];
			index = (++index) % tableSize;
		}
		return NULL;
	}
	~EdgeTable() {
		delete[] data;
	}
};


/*******************************************************
 * 
 * The Mesh class
 * ==============
 *
 * This is the main class of this application. It constains
 * all the data for a mesh and defines the mesh manipulation
 * methods, most notably the "subdivide" method, which subdivides
 * a mesh and returns a new subdivided version.
 *
 * In this implementation of a Mesh, I make use of an associative
 * store of Edge records (the "EdgeTable" type) rather than
 * using a more tightly-linked structure. My past experience
 * is that the tightly-linked structures are very hard to manipulate.
 *
 * A mesh consists of:
 *
 * (1) A table of all the vertices in the mesh (an STL vector).
 * (2) A table of all the faces in the mesh (also an STL vector)
 * (3) A table of all the edges in the mesh (a hash table). Note
 *     that each edge appears twice, once for each "direction".
 *
 * The vertex table and face table are both indexable like arrays.
 *
 * A mesh is built by first defining all the vertices, then
 * adding faces one by one, each face being just a list of
 * vertex indices. The edge table is built as each face is added.
 *
 *******************************************************/
class Mesh {
  public:
   FaceList faces;
   EdgeTable* edges;
   PointList vertices;
   friend ostream& operator<<(ostream& s, Mesh& vf);
   
   Mesh(int maxNumEdges) {
   // Constructs a new empty mesh.
	   edges = new EdgeTable(maxNumEdges);
   }
   
   Mesh(int numVertices, CVec3df vertices[], int faceIndices[]) {
      // Constructs a new mesh from the given array of vertices and the
      // given array of faces. The latter is an integer array in which
      // each face is defined as a sequence of indices into the vertices
      // array, with each face terminated by the value -1. The whole list
      // is terminated by -1 (i.e. an empty face list).
      edges = new EdgeTable(10 * numVertices);
      for (int i = 0; i < numVertices; i++)
         addVertex(vertices[i]);
      Face* newFace = 0;
      for (int i = 0; ; i++) {
            if (newFace == 0) newFace = new Face();
            if (faceIndices[i] == -1) {
                if (newFace->size() != 0) {
                    addFace(*newFace);
                    delete newFace;
                    newFace = 0;
                }
                else {
                    delete newFace;
                    return;
                }
            } else 
                newFace->addVertex(faceIndices[i]);
      }
   }
         
   
   int addVertex(CVec3df v) {
   // Adds a vertex to a mesh. The return value is the 
   // ID (index) of the newly inserted vertex.
      vertices.push_back(v);
      return (int) (vertices.size() - 1);
   }
   
   int addFace(Face f) {
   // Adds a face to a mesh. The return value is the ID (index) of the newly
   // inserted face. Also adds entries to the edge table for each edge.
      int faceID = (int) faces.size();   // This is the new face ID (i.e. index into face table) 
      faces.push_back(f);
      int n = (int) f.size();
      for (int vx = 0; vx < n; vx++) { // For each edge
         int v1 = f[vx];               // Edge start vertex
         int v2 = f[(vx+1) % n];       // Edge end vertex
         Edge e(v1, v2, faceID);        // Construct the edge record
         edges->insert(e);
      }
      
      return faceID;
   }
      
   CVec3df newFaceVertex(int faceNum, int vertexNum) {
        // Computes the location of a new vertex in face "faceNum", vertex
        // "vertexNum". This is just a matter of computing the appropriately
        // weighted sum of the face vertices using the Doo-Sabin weights.
        static float quadWeights[] = {9.0f/16, 3.0f/16, 1.0f/16, 3.0f/16};
        Face f = faces[faceNum];
        int n = f.size();
        CVec3df sum(0,0,0);
        for (int i = 0; i < n; i++) {
            float weight;
            if (n == 4) weight = quadWeights[i];
            else if (i == 0) weight = 1.0f/4 + 5.0f/(4 * n);
            else weight = (3 + 2 * cos(2 * i * PI / n))/(4 * n);
            sum += vertices[f[(vertexNum + i) % n]] * weight ;
        }
        return sum;
    };

    int getLeftFace(Face& face, int vertIndex) {
    // Returns the index to the face that is on the left of the given input
    // face when seen from the given vertIndex (the index within the given face).
    // Method is to find the edge in this face that starts on the given
    // vertex when going clockwise, and to take the left face of that edge.
        
        int n = face.size();
        int previousVertexIndex = (vertIndex + n - 1) % n;
        Edge e(face[vertIndex], face[previousVertexIndex], 0);
        Edge* eInTable = edges->find(e);
        assert (eInTable != NULL);
        return eInTable->leftFace;
    }
    
    // The "main" mesh method -- subdivide
    // ===================================
    
    Mesh* subdivide() {
      // Returns a pointer to a new mesh that is the subdivided version
      // of this. The new mesh is on the heap and the client must delete it
      // when it is no longer needed.
      
	  int maxNumEdges = (int) faces.size() * 4 * 6;  // 4 new vertices / face, max valence = 6 [guess]
      Mesh* newMesh = new Mesh( maxNumEdges ); // An empty mesh.
      
      // Firstly, loop through all faces of the mesh.
      // For each face, generate new vertices and construct the new internal face.
      // As each new face vertex is created it is stored in an array of
      // vectors, one per face.
      
      intList* childVertices = new intList[faces.size()];  // One vector of new vertices for each face
      for (int i = 0; i < (int) faces.size(); i++ ) {
            Face newFace;  // The new internal face we're constructing for this old face
            for (int j = 0; j < (int) faces[i].size(); j++) {
               CVec3df newVert = newFaceVertex(i,j);  // Construct the new vertex in the face
               int newVertIndex = newMesh->addVertex(newVert);  // Add it to the new mesh
               childVertices[i].push_back(newVertIndex);  // Save where it went for later use
               newFace.addVertex(newVertIndex);    // Add it to the new face         
            }
            newMesh->addFace(newFace);             // Add the face to the new mesh
      }
      
      // Now create all the new edge faces by iterating over the edges.

      for (Edge* ep = edges->firstEdge; ep != NULL; ep = ep->next) {
          if (ep->v2 > ep->v1) {  // Each edge appears twice -- only deal with it once
              Edge eOpposite(ep->v2,ep->v1,0);
              Edge* eOpp = edges->find(eOpposite);  // Get the opposing edge
              assert (eOpp != 0);
              int lfx = ep->leftFace;     // The index (ID) of the edge's left face
              int rfx = eOpp->leftFace;   // The index of the edge's right face
              Face& fLeft = faces[lfx];   // The actual left face
              Face& fRight = faces[rfx];  // And the right face
              int leftOffset = fLeft.indexToVertex(ep->v1); // Find the edge's start vertex in the left face
              int rightOffset = fRight.indexToVertex(ep->v1);  // ... and in the right face
              assert (leftOffset > -1 && rightOffset > -1);
              int v1 = childVertices[lfx][leftOffset];  // Lookup the ...
              int v2 = childVertices[rfx][rightOffset]; // ... vertices of the new edge face.
              int v3 = childVertices[rfx][(rightOffset + fRight.size() - 1) % fRight.size()];
              int v4 = childVertices[lfx][(leftOffset + 1) % fLeft.size()];
              int verts[] = {v1, v2, v3, v4};
              Face newFace(4, verts);
              newMesh->addFace(newFace); 
           }         
      } 

      // Lastly process all the new vertex faces.
      // Do this by processing all the faces, keeping a boolean table of
      // processed vertices.

      bool* vertexDone = new bool[vertices.size()];
      for (int i = 0; i < (int) vertices.size(); i++) vertexDone[i] = false;
            
      for (int faceIndex = 0; faceIndex < (int) faces.size(); faceIndex++ ) {
          Face& face = faces[faceIndex]; // The current face
          for (int vx = 0; vx < face.size(); vx++) { // Try each vertex of the face
             int vertex = face[vx];      // Get the actual vertex ID (index)
             if (!vertexDone[vertex]) {  // If not already processed ...
                 Face newFace;           // Start creating a new face
                 int thisFaceIndex = faceIndex;
                 int checkCounter = 0;   // Debug -- catch infinite loops
                 // Construct the face by looping around all faces containing this vertex
                 do {
                     Face& thisFace = faces[thisFaceIndex];
                     int indexInThisFace = thisFace.indexToVertex(vertex);
                     assert(indexInThisFace > -1);
                     newFace.addVertex(childVertices[thisFaceIndex][indexInThisFace]);
                     thisFaceIndex = getLeftFace(thisFace, indexInThisFace);
                     assert (++checkCounter < 10);
                 } while (thisFaceIndex != faceIndex && checkCounter < 10);
                 newMesh->addFace(newFace);
                 vertexDone[vertex] = true;                   
             }
          }
      }
	  delete[] vertexDone;
	  delete[] childVertices;
      cout << "Subdivision done, num faces = " << (int) newMesh->faces.size() << endl;                           
      return newMesh;
   } 
   
   CVec3df* getFace(int faceNum, int& numVertices) { 
      // Returns the 'faceNum'th face as an array of vertices.
      // The array is allocated on the heap; the user is responsible for
      // deleting it.
      // Ref parameter numVertices is set to the number of vertices.
      Face face = faces[faceNum];
      numVertices = face.size();
      CVec3df* points = new CVec3df[numVertices];
      for (int i = 0; i < numVertices; i++) {
         points[i] = vertices[face[i]];
      }
      return points;
   }

   ~Mesh() {
	   delete edges;
   }
};

ostream& operator<<(ostream& s, Mesh& mesh) {
    // This is just a debug routine, with a very odd output format.
    // I couldn't quite bring myself to throw it away, so it's still here.
     int numFaces = (int) mesh.faces.size();
     s << "Face has " << numFaces << " face\n";
     for (int i = 0; i < numFaces; i++) {
        Face f = mesh.faces[i];
        s << "Face " << i << ":\n" << f << endl;
        int numVertices;
        CVec3df* vertices = mesh.getFace(i, numVertices);
        for (int j = 0; j < numVertices; j++)
           s << vertices[j] << endl;
     }
     s << endl << flush;
     return s;
}

/*******************************************************************
 *
 * The "main program" stuff starts here
 *
 *******************************************************************/    
   

CTrackball trackball;

CVec3df cubeVertices[] = { CVec3df(-1,-1,-1), CVec3df(1,-1,-1), CVec3df(-1,1,-1),
    CVec3df(1,1,-1), CVec3df(-1,-1,1), CVec3df(1,-1,1), CVec3df(-1,1,1),CVec3df(1,1,1)};
    
int faceIndices[] = {
    6,7,3,2,-1,  // top
    0,2,3,1,-1,  // back
    1,3,7,5,-1,  // right
    2,0,4,6,-1,  // left
    0,1,5,4,-1,  // bottom
    4,5,7,6,-1,-1}; // front
    
  
Mesh cube(8,cubeVertices,faceIndices);

Mesh* currentMesh = &cube;
const int MAX_SUBDIVISIONS = 5;
const int MAX_DIVISIONS_PLUS_1 = 6;
Mesh** savedMeshes = new Mesh*[MAX_DIVISIONS_PLUS_1];
int subdivisionLevel = 0;

void display(void) {
     glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
     glMatrixMode(GL_PROJECTION);
     glLoadIdentity();
     gluPerspective(40,
             glutGet(GLUT_WINDOW_WIDTH) / (float) glutGet(GLUT_WINDOW_HEIGHT),
             0.1, 10);
     glMatrixMode(GL_MODELVIEW);
     glLoadIdentity();
     gluLookAt(0, 0, 6, 0, 0, 0, 0, 1, 0);
     glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
     trackball.tbMatrix();
     glColor3f(0,0,0);     // Draw cube in black
     
     for (int i = 0; i < (int) currentMesh->faces.size(); i++) {
          int numVerts;
          CVec3df* points = currentMesh->getFace(i, numVerts);
          glBegin(GL_POLYGON);
          for (int j = 0; j < numVerts; j++) {            
             glVertex3fv(points[j].getArray());
          }
          glEnd();
          delete[] points;
     }
          
     glFlush();
     glutSwapBuffers();
}

void handleMouseMotion(int x, int y) {   
    trackball.tbMotion(x, y);
}

void handleMouseClick(int button, int state, int x, int y) {
    trackball.tbMouse(button, state, x, y);
}

void handleKeyboardEvent(unsigned char key, int x, int y) {
    if ((key == '+' || key == '=') && subdivisionLevel < MAX_SUBDIVISIONS) {
        subdivisionLevel++;
        if (savedMeshes[subdivisionLevel] == 0)
            savedMeshes[subdivisionLevel] = currentMesh->subdivide();
        currentMesh = savedMeshes[subdivisionLevel];
    }
    else if ((key == '-' || key == '_') && subdivisionLevel > 0) {
        subdivisionLevel--;
        currentMesh = savedMeshes[subdivisionLevel];
    }
    else 
        trackball.tbKeyboard(key);
    glutPostRedisplay();
}

void reshape(int width, int height ) {
    // Called at start, and whenever user resizes component
    trackball.tbReshape(width, height);
    glViewport(0, 0, width, height);
}


int main(int argc, char** argv) {
     for (int i = 0; i <= MAX_SUBDIVISIONS; i++) savedMeshes[i] = 0;
     savedMeshes[0] = &cube; 
     trackball.tbInit(GLUT_LEFT_BUTTON);
     glutInit(&argc, argv);
     glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB);
     glutInitWindowSize(400,400);
     glutInitWindowPosition(50,50);
     glutCreateWindow("Subdivision");
     glutMouseFunc(handleMouseClick);        // Set function to handle mouse clicks
     glutMotionFunc(handleMouseMotion);      // Set function to handle mouse motion
     glutKeyboardFunc(handleKeyboardEvent);  // Set function to handle keyboard input
     glutReshapeFunc(reshape);
     glutDisplayFunc(display);
     glClearColor(1,1,1,1);
     glEnable(GL_DEPTH_TEST);
     glutMainLoop();
     return(0);
}


